subroutine gray_jetto1beam(ijetto, mr, mz, r, z, psin, psia, rax, zax, &
    nbnd, rbnd, zbnd, nrho, psrad, fpol, te, dne, zeff, qpsi, ibeam,   &
    p0mw, alphain, betain, dpdv, jcd, pabs, icd, ierr)
  use const_and_precisions, only : wp_
  use units, only : ucenr,usumm,uprj0,uwbm,udisp,ubres,ucnt,uoutr,ueq,uprfin,  &
      uflx,upec,uprm,ubeam
  use gray_params, only : read_params,rtrparam_type,hcdparam_type,antctrl_type,&
      eqparam_type,prfparam_type,outparam_type
  use beams, only : read_beam2
  use graycore, only : gray_main
  use reflections, only : range2rect
  use coreprofiles, only : tene_scal
  implicit none
! arguments
  integer, intent(in) :: ijetto, mr, mz, nbnd, nrho, ibeam
  real(wp_), dimension(mr), intent(in) :: r
  real(wp_), dimension(mz), intent(in) :: z
  real(wp_), dimension(mr,mz), intent(in) :: psin
  real(wp_), intent(in) :: psia, rax, zax, p0mw, alphain, betain
  real(wp_), dimension(nbnd), intent(in) :: rbnd, zbnd
  real(wp_), dimension(nrho), intent(in) :: psrad, fpol, te, dne, zeff, qpsi
  real(wp_), dimension(nrho), intent(out) :: dpdv, jcd
  real(wp_), intent(out) :: pabs, icd
  integer, intent(out) :: ierr
! local variables
  real(wp_), dimension(nrho) :: psinr
  integer :: iox0
  real(wp_) :: r0m,rvac,alpha0,beta0,psipol0,chipol0,rwallm
  real(wp_) :: fghz,x0,y0,z0,w1,w2,ri1,ri2,phiw,phir
  real(wp_), dimension(5) :: rlim,zlim
  logical, save :: firstcall=.true.
  type(rtrparam_type) :: rtrp
  type(hcdparam_type) :: hcdp
  type(antctrl_type) :: antp
  type(eqparam_type) :: eqp
  type(prfparam_type) :: prfp
  type(outparam_type) :: outp

! if first call read parameters from external file
  if (firstcall) then
    call read_params('gray.data',rtrp,hcdp,antp,eqp,rwallm,prfp,outp,uprm)
    rwallm=r(1)
    antp%filenm='graybeam.data'
    eqp%filenm='JETTO'
    eqp%iequil=ijetto+1
    prfp%filenm='JETTO'
    outp%ipec=1
    firstcall=.false.
    outp%nrho=nrho
  end if
  
!  call tene_scal(te,dne,prfp%factte,prfp%factne,eqp%factb,&
!                 prfp%iscal,prfp%iprof)
  
  rvac=rax
  psinr=psrad

  alpha0=alphain
  beta0=betain
  call read_beam2(antp%filenm,ibeam,alpha0,beta0,fghz,antp%iox,x0,y0,z0, &
      w1,w2,ri1,ri2,phiw,phir,ubeam)
  psipol0=antp%psi
  chipol0=antp%chi
  iox0=antp%iox

! set simple limiter
  r0m=sqrt(x0**2+y0**2)*0.01_wp_
  call range2rect(rwallm,max(r0m,r(mr)),z(1),z(mz),rlim,zlim)

! call main subroutine for the ibeam-th beam
  call gray_main(r,z,psin,psia,psinr,fpol,qpsi,rvac,rax,zax,rbnd,zbnd,eqp,     &
      psrad,te,dne,zeff,prfp, rlim,zlim, p0mw,fghz,alpha0,beta0,(/x0,y0,z0/),  &
      w1,w2,ri1,ri2,phiw,phir,iox0,psipol0,chipol0, dpdv,jcd,pabs,icd,outp,    &
      rtrp,hcdp,ierr,sqrt(psrad))

! close output (debug) files
  close(ucenr)
  close(usumm)
  close(uprj0)
  close(uprj0+1)
  close(uwbm)
  close(udisp)
  close(ubres)
  close(ucnt)
  close(uoutr)
  close(ueq)
  close(uprfin)
  close(uflx)
  close(upec)

end subroutine gray_jetto1beam
